/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;

/**
 * @author Gilles Bernard
 * Audio thread.
 * contantly read from the default input line a forward buffer through a ISampleListener interface.
 * The number of sample to be read control how many sample are read and forwarded to the sample listener.
 * It thus control the rate of processing and accuracy of signal analysis. 
 */
public class AudioThread extends Thread {
	private TargetDataLine line=null;
	private byte[] buffer = null;
	private Vector<ISampleListener> listeners;
	private CyclingBuffer cyclingBuffer = new CyclingBuffer(Config.globalBufferSize()); 
	private boolean pause=false;
	private int nbSamples;
	private short[] sample;
	private float sampleRate;

	public AudioThread(Vector<ISampleListener> listeners, int nbSamples) throws LineUnavailableException {

		init(Config.defaultSampleRate());

		this.buffer=new byte[nbSamples*Config.sampleSizeInBytes()];
		this.sample=new short[nbSamples];
		this.listeners=listeners;
		this.nbSamples=nbSamples;
		this.setName("AudioThread");
	}

	private synchronized void init(float sampleRate) throws LineUnavailableException {
		if(line!=null) line.stop();

		int sampleSizeInBits = Config.sampleSizeInBits();
		int channels = 1;
		boolean signed = true;
		boolean bigEndian = true;
		this.sampleRate=sampleRate;
		AudioFormat format = new AudioFormat(sampleRate, sampleSizeInBits, channels, signed,
				bigEndian);
		DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);
		line = (TargetDataLine) AudioSystem.getLine(info);
		line.open(format);
		line.start();
	}

	public void run() {
		int count=0;
		while(!isInterrupted()) {
			synchronized(this) {
				if(buffer==null || buffer.length!=nbSamples) {
					this.buffer=new byte[nbSamples*Config.sampleSizeInBytes()];
					this.sample=new short[nbSamples];
				}

				count=line.read(buffer, 0, buffer.length);
			}
			if(count==buffer.length && !pause) {
				ByteBuffer.wrap(buffer).order(ByteOrder.BIG_ENDIAN).asShortBuffer().get(sample);
				cyclingBuffer.put(sample);
				for(ISampleListener listener : listeners) {
					listener.process(sample);
				}
			}
		}

	}

	public void setNbSamples(int nbSamples) {
		this.nbSamples=nbSamples;
	}

	public void setPause(boolean selected) {
		pause=selected;
	}

	public void setSampleRate(float f) {
		try {
			init(f);
		} catch (LineUnavailableException e) {
			e.printStackTrace();
		}
	}

	public float getSampleRate() {
		return sampleRate;
	}

	public int getNbSamples() {
		return nbSamples;
	}


}
