package com.lordikc.audioanalyst;

import java.util.Arrays;

/**
 * @author Gilles Bernard
 * Cycling buffer allowing a continuous write and giving read access to previous values stored in the cycling buffer.
 * Writing process starts back from the beginning when end is reached overwriting values older than the buffer size.
 */
public class CyclingBuffer {

	private short[] buffer;
	private int idxWrite;

	/**
	 * @param size
	 */
	public CyclingBuffer(int size) {
		buffer=new short[size];
		clear();
	}

	/**
	 * Clear the buffer and set writing index to 0.
	 */
	public synchronized void clear() {
		idxWrite=0;
		Arrays.fill(buffer, (short)0);
	}
	/**
	 * Add sample to the buffer starting back from the beginning if end is reached.
	 * @param sample
	 */
	public synchronized void put(short[] sample){
		if((idxWrite+sample.length)>=this.buffer.length) {
			int towrite=this.buffer.length-idxWrite; 
			if(towrite>0)
				System.arraycopy(sample, 0, this.buffer, idxWrite, towrite);
			if((sample.length-towrite)>0)
				System.arraycopy(sample, towrite, this.buffer, 0, sample.length-towrite);
			idxWrite+=sample.length;
			idxWrite%=this.buffer.length;
		} else {
			System.arraycopy(sample, 0, this.buffer, idxWrite, sample.length);
			idxWrite+=sample.length;
		}
	}

	/**
	 * Add value to the buffer starting back from the beginning if end is reached.
	 * @param val
	 */
	public synchronized void put(short val){
		buffer[idxWrite]=val;
		idxWrite++;
		idxWrite%=buffer.length;
	}

	/**
	 * Returns the size of the buffer
	 * @return
	 */
	public int size() { return buffer.length;}
	
	
	/**
	 * Returns the current write index.
	 * @return
	 */
	public int getIndex() { return idxWrite;}
		
	/**
	 * Fill the sample with data starting at the given offset.
	 * If offset is negative, use current position minus |offset|
	 * @param sample
	 * @param offset
	 */
	public synchronized void getData(short[] sample, int offset) {
		if(offset<0) {
			offset+=idxWrite;
			if(offset<0)
				offset+=buffer.length;
		}
		if(offset>=buffer.length) offset%=buffer.length;
		if((sample.length+offset)>=buffer.length) {
			int towrite=this.buffer.length-offset; 
			if(towrite>0)
				System.arraycopy(this.buffer, offset, sample, 0, towrite);
			if((sample.length-towrite)>0)
				System.arraycopy(this.buffer, 0, sample, towrite, sample.length-towrite);
		} else {
			System.arraycopy(buffer, offset, sample, 0, sample.length);
		}
	}


}
