/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.HierarchyBoundsListener;
import java.awt.event.HierarchyEvent;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * @author Gilles Bernard
 * Main class for the Enveloppe. It extends the JFrame class and control the audio and processing threads.
 */
public class Enveloppe extends JFrame implements ChangeListener, ActionListener, HierarchyBoundsListener {

	private static final long serialVersionUID = 5312114580592842412L;
	private EnveloppePanel panel;
	private AudioThread audioThread=null;
	private JPanel command;
	private JCheckBox memCheckBox;
	private JCheckBox dbCheckBox;
	private VolumePanel volumePanel;
	private JSlider scaleSlider;
	private JLabel scaleLabel;

	public Enveloppe() {
		setTitle("Lordikc AudioAnalyst Enveloppe");
		InputStream is = getClass().getResourceAsStream("/oscillo_128.png");
		try {
			if(is!=null)
				this.setIconImage(ImageIO.read(is));
			else System.out.println("Can't load icon");
		} catch (IOException e) {
			e.printStackTrace();
		}
		setLayout(new BorderLayout());
		// The panel display the chart
		panel = new EnveloppePanel();
		getContentPane().add(panel,BorderLayout.CENTER);
		// The command panel
		command = new JPanel();
		getContentPane().add(command,BorderLayout.EAST);
		command.add(Box.createRigidArea(new Dimension(0,5)));
		command.setLayout(new BoxLayout(command, BoxLayout.PAGE_AXIS));

		// Horizontal time scale label, slider and value
		JLabel label=new JLabel("Time Scale");
		command.add(label);
		scaleSlider = new JSlider(JSlider.HORIZONTAL,
				0, Config.scalesSpectrogram().length-1, Config.defaultScaleSpectrogram());
		scaleSlider.addChangeListener(this);
		scaleSlider.setPreferredSize(new Dimension(100,20));
		command.add(scaleSlider);
		scaleLabel=new JLabel();
		command.add(scaleLabel);

		// Hold the display
		memCheckBox=new JCheckBox("Hold");
		memCheckBox.setActionCommand(Config.HOLD);
		memCheckBox.addActionListener(this);
		command.add(memCheckBox);

		// Select dB or decimal display
		dbCheckBox=new JCheckBox("dB");
		dbCheckBox.setActionCommand(Config.DB);
		dbCheckBox.addActionListener(this);
		command.add(dbCheckBox);

		// About button
		JButton button=new JButton("About");
		button.setActionCommand("about");
		button.addActionListener(this);
		command.add(button);

		// Volume panel
		volumePanel=new VolumePanel();
		Box bv = Box.createVerticalBox();
		bv.add(Box.createVerticalGlue());
		bv.add(volumePanel);
		bv.add(Box.createVerticalGlue());
		Box bh = Box.createVerticalBox();
		bh.add(Box.createHorizontalGlue());
		bh.add(bv);
		bh.add(Box.createHorizontalGlue());
		command.add(bh);

		this.getContentPane().addHierarchyBoundsListener(this);

		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		pack();
	}

	public void init() {
		setScale(Config.scalesSpectrogram()[Config.defaultScaleSpectrogram()]);
	}

	private void setScale(double d) {
		scaleLabel.setText(Math.round(d)+"s");
		panel.setScale(d);
		// Compute nb of samples to be used
		if(getWidth()>0) {
			int nbSamples = (int) (d*Config.defaultSampleRate()/getWidth());
			if(audioThread!=null && nbSamples<=Config.defaultSampleRate()) {
				audioThread.setNbSamples(nbSamples);
			}
		}
	}

	public static void main(String args[]) throws Exception {

		// Create the main frame and attach the audio thread to it
		final Enveloppe frame = new Enveloppe();
		EnveloppeProcessThread process = new EnveloppeProcessThread(1<<Config.defaultSamples(),frame);
		Vector<ISampleListener> listeners = new Vector<ISampleListener>();
		listeners.add(process);
		AudioThread audio = new AudioThread(listeners,1<<Config.defaultSamples());
		frame.setAudioThread(audio);
		frame.init();
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				frame.setVisible(true);
			}
		});

		// Start the threads
		process.start();
		audio.start();

		// And wait for completion
		audio.join();
		process.join();
	}

	private void setAudioThread(AudioThread audio) {
		this.audioThread=audio;
	}

	@Override
	public void stateChanged(ChangeEvent event) {
		if(event.getSource().equals(scaleSlider)) {
			setScale(Config.scalesSpectrogram()[scaleSlider.getValue()]);
			panel.repaint();
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		if(event.getActionCommand().equals(Config.HOLD)){
			audioThread.setPause(memCheckBox.isSelected());
			panel.setPause(memCheckBox.isSelected());
			panel.repaint();
		} else if(event.getActionCommand().equals(Config.DB)){
			panel.setDBDisplay(dbCheckBox.isSelected());
			panel.repaint();
		} else if(event.getActionCommand().equals(Config.ABOUT)){
			Util.about(this);
		}
	}

	public void setVolume(short volume) {
		volumePanel.setVolume(volume);
		volumePanel.repaint();
		panel.setVolume(volume);
		panel.repaint();
	}

	@Override
	public void ancestorMoved(HierarchyEvent arg0) {
	}

	@Override
	public void ancestorResized(HierarchyEvent e) {
		panel.reset();
		panel.repaint();
	}

}

