/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.util.Arrays;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.border.BevelBorder;

/**
 * @author Gilles Bernard
 * Panel for displaying the Enveloppe. X is time, Y is frequency and value is represented by a color scale.
 */
public class EnveloppePanel extends JPanel {

	private static final long serialVersionUID = 6360887160294401903L;


	private double scale; // Duration for the panel width in second
	private short[] volumes=null;
	private long startTime=0;
	private long pauseTime=0;
	private double duration;


	private boolean dbDisplay=false;


	private int lastIdx=0;

	public EnveloppePanel() {
		setPreferredSize(new Dimension(800,600));
		setBackground(new java.awt.Color(255, 255, 255));
		setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));	
		reset();
	}

	/* (non-Javadoc)
	 * @see javax.swing.JComponent#paint(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g){
		try {
			// Reset the display
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, getWidth(), getHeight());

			// Draw the chart
			// Current index in volumes
			long time=Util.clock();
			time-=startTime;
			time*=volumes.length;
			time/=duration;
			time%=volumes.length;

			// This index must be placed on the right side of the panel
			g.setColor(Color.RED);
			for(int i=0;i<getWidth();i++)
			{
				int v=volumes[(int) ((i+time+getWidth())%volumes.length)];
				if(v>=0) {
					int y=0;
					if(dbDisplay) {
						double att=Config.MIN_DB;
						if(v>0){
							att=10.*Math.log10((double)v/(double)Short.MAX_VALUE);
							if(att<Config.MIN_DB) att=Config.MIN_DB;
							y=(int) (att*getHeight()/Config.MIN_DB);
						} else y=getHeight();
					}else{
						y=getHeight()-(v*getHeight())/Short.MAX_VALUE;
					}

					g.drawLine(i,getHeight(),i,y);
				}
			}

			// Draw horizontal grid
			for(int i=1;i<Config.nbGridsY();i++){
				if(i==Config.nbGridsY()/2)
					g.setColor(Color.BLACK);
				else
					g.setColor(Color.GRAY);
				int row=(this.getHeight()*i)/Config.nbGridsY();
				g.drawLine(0, row, this.getWidth(), row);
			}


			// Draw vertical grid
			for(int i=1;i<Config.nbGridsX();i++) {
				if(i==Config.nbGridsX()/2)
					g.setColor(Color.BLACK);
				else
					g.setColor(Color.GRAY);
				int col=(getWidth()*i)/Config.nbGridsX();
				g.drawLine(col, 0, col, this.getHeight());
			}

		} catch (Exception e){
			e.printStackTrace();
		}
	}

	/**
	 * Set the current volume as a positive short value. 
	 * @param volume
	 */
	public void setVolume(short volume) {
		long time=Util.clock();
		time-=startTime;
		time*=volumes.length;
		time/=duration;
		int idx=((int)time)%volumes.length;
		volumes[idx]=volume;
		if(idx>lastIdx) {
			for(int i=lastIdx+1;i<idx;i++)
				volumes[i]=volume;
		} else if (idx<lastIdx) {
			for(int i=lastIdx+1;i<(idx+volumes.length);i++)
				volumes[i%volumes.length]=volume;
		}
		lastIdx=idx;
	}

	/**
	 * Reset the display. 
	 */
	public void reset() {
		if(getHeight()>0 && getWidth()>0) {
			volumes=new short[getWidth()];
			Arrays.fill(volumes, (short)0);
			startTime=Util.clock();
			pauseTime=0;
			lastIdx=0;
			// s duration of the whole screen
			duration=scale*Util.clock_to_second();
		}
	}


	/**
	 * Set the vertical scale. The scale is the number of second for the whole panel width.
	 * @param d
	 */
	public void setScale(double d) {
		this.scale=d;
		reset();
	}

	/**
	 * Called whenever the pause is selected/unselected.
	 * Remove the pause time so the display stay consistent
	 * @param selected
	 */
	public void setPause(boolean selected) {
		if(!selected) {
			long timeCur=Util.clock();
			startTime+=timeCur-pauseTime;
			pauseTime=0;
		} else {
			pauseTime=Util.clock();
		}
	}

	public void setDBDisplay(boolean selected) {
		this.dbDisplay=selected;
	}


}
