/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * @author Gilles Bernard
 * Main class of the oscilloscope. The oscilloscope display the amplitude in Y and the time in X.
 * The display can be synchronized on a up or down transition of the signal.
 * The class overrides JFrame and is the main frame of the application.
 */
public class Oscilloscope extends JFrame implements ChangeListener, ActionListener {

	private static final long serialVersionUID = 5312114580592842412L;
	private static int nbSamples = 1<<14;
	private OscilloscopePanel panel;
	private AudioThread audioThread=null;
	private JPanel command;
	private JSlider scaleSlider;
	private JLabel scaleLabel;
	private JCheckBox upDownCheckBox;
	private JSlider thresholdSlider;
	private JLabel thresholdLabel;
	private JCheckBox memCheckBox;
	private VolumePanel volumePanel;
	private JSlider nbSampleSlider;
	private JLabel nbSampleLabel;

	public Oscilloscope() {
		setTitle("Lordikc AudioAnalyst Oscilloscope");
		InputStream is = getClass().getResourceAsStream("/oscillo_128.png");
		try {
			if(is!=null)
				this.setIconImage(ImageIO.read(is));
			else System.out.println("Can't load icon");
		} catch (IOException e) {
			e.printStackTrace();
		}
		setLayout(new BorderLayout());
		// The panel display the chart
		panel = new OscilloscopePanel();
		getContentPane().add(panel,BorderLayout.CENTER);
		// The command panel
		command = new JPanel();
		getContentPane().add(command,BorderLayout.EAST);
		command.add(Box.createRigidArea(new Dimension(0,5)));
		command.setLayout(new BoxLayout(command, BoxLayout.PAGE_AXIS));

		// 0.5s -> 0.00005s
		scaleSlider = new JSlider(JSlider.HORIZONTAL,
				0, Config.timeScales().length-1, Config.defaultTimeScale());
		scaleSlider.addChangeListener(this);
		scaleSlider.setPreferredSize(new Dimension(100,20));
		JLabel label=new JLabel("Time Scale");
		command.add(label);
		command.add(scaleSlider);
		scaleLabel=new JLabel();
		command.add(scaleLabel);
		setScale(Config.timeScales()[Config.defaultTimeScale()]);

		// Nb samples per second
		nbSampleSlider = new JSlider(JSlider.HORIZONTAL,
				Config.minSamples(), Config.maxSamples(), Config.defaultSamples());
		nbSampleSlider.addChangeListener(this);
		nbSampleSlider.setPreferredSize(new Dimension(100,20));
		label=new JLabel("Samples/scan");
		command.add(label);
		command.add(nbSampleSlider);
		nbSampleLabel=new JLabel();
		command.add(nbSampleLabel);

		upDownCheckBox = new JCheckBox("Syn. U/D");
		upDownCheckBox.setActionCommand(Config.UPDOWNFRONT);
		upDownCheckBox.addActionListener(this);
		command.add(upDownCheckBox);
		// 0% 100%
		thresholdSlider = new JSlider(JSlider.HORIZONTAL,
				0, 20, 3);
		thresholdSlider.setPreferredSize(new Dimension(100,20));

		thresholdSlider.addChangeListener(this);
		label=new JLabel("Threshold");
		command.add(label);
		command.add(thresholdSlider);
		thresholdLabel=new JLabel();
		command.add(thresholdLabel);
		setThreshold((thresholdSlider.getValue()*100)/thresholdSlider.getMaximum());

		memCheckBox=new JCheckBox("Hold");
		memCheckBox.setActionCommand(Config.HOLD);
		memCheckBox.addActionListener(this);
		command.add(memCheckBox);

		JButton button=new JButton("About");
		button.setActionCommand("about");
		button.addActionListener(this);
		command.add(button);

		volumePanel=new VolumePanel();
		Box bv = Box.createVerticalBox();
		bv.add(Box.createVerticalGlue());
		bv.add(volumePanel);
		bv.add(Box.createVerticalGlue());
		Box bh = Box.createVerticalBox();
		bh.add(Box.createHorizontalGlue());
		bh.add(bv);
		bh.add(Box.createHorizontalGlue());
		command.add(bh);

		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		pack();
	}

	public void init() {
		setNbSamples(1<<nbSampleSlider.getValue());
	}


	/**
	 * Sets the horizontal scale corresponding to the duration of the width of the panel.
	 * @param d
	 */
	private void setScale(double d) {
		panel.setScale(d);
		scaleLabel.setText(Util.periodToText(d));
		int nbSamples=(int) (d*Config.defaultSampleRate());
		int nbSamples2=(int)Util.next_power_of_2(nbSamples);
		nbSamples2*=2;

		int min=(int)Util.next_power_of_2(getWidth())*2;
		if(nbSamples2<min) nbSamples2=min;

		if(audioThread!=null) audioThread.setNbSamples(nbSamples2);
	}

	private void setNbSamples(int i) {
		if(audioThread!=null)
			audioThread.setNbSamples(i);
		nbSampleLabel.setText(Integer.toString(i));
	}

	public static void main(String args[]) throws Exception {

		final Oscilloscope frame = new Oscilloscope();
		OscilloscopeProcessThread process = new OscilloscopeProcessThread(nbSamples,frame);
		Vector<ISampleListener> listeners = new Vector<ISampleListener>();
		listeners.add(process);
		AudioThread audio = new AudioThread(listeners,nbSamples);
		frame.setAudioThread(audio);
		frame.init();
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				frame.setVisible(true);
			}
		});

		process.start();
		audio.start();

		audio.join();
		process.join();
	}

	private void setAudioThread(AudioThread audio) {
		this.audioThread=audio;
	}

	public void setBuffer(short[] buffer, short minValue, short maxValue) {
		if(!panel.isPause()) {
			panel.setBuffer(buffer, minValue, maxValue);
			volumePanel.setVolume(AudioProcessing.volume(buffer));
			panel.repaint();
			volumePanel.repaint();
		}
	}

	@Override
	public void stateChanged(ChangeEvent event) {
		if(event.getSource().equals(scaleSlider)) {
			setScale(Config.timeScales()[scaleSlider.getValue()]);
			panel.setPause(false);
			audioThread.setPause(false);
			panel.repaint();
			memCheckBox.setSelected(false);
		}
		if(event.getSource().equals(thresholdSlider)) {
			setThreshold((thresholdSlider.getValue()*100)/thresholdSlider.getMaximum());
		}
		if(event.getSource().equals(nbSampleSlider)) {
			setNbSamples(1<<nbSampleSlider.getValue());
			panel.repaint();
		}
	}

	private void setThreshold(float d) {
		panel.setThreshold(d);
		panel.repaint();
		thresholdLabel.setText(Integer.toString((int)d)+"%");
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		if(event.getActionCommand().equals(Config.HOLD)){
			panel.setPause(memCheckBox.isSelected());
			audioThread.setPause(memCheckBox.isSelected());
			panel.repaint();
		}
		if(event.getActionCommand().equals(Config.UPDOWNFRONT)){
			panel.setSynchro(upDownCheckBox.isSelected());
			panel.repaint();
		} else if(event.getActionCommand().equals(Config.ABOUT)){
			Util.about(this);
		}
	}



}

