/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.border.BevelBorder;

/**
 * @author Gilles Bernard
 * Main panel for displaying the chart of the oscilloscope.
 */
public class OscilloscopePanel extends JPanel {

	private static final long serialVersionUID = 455525426556772083L;
	private short[] buffer=null;
	private short min;
	private short max;
	private int index;
	private double scale;
	private float threshold;
	private boolean synchro;
	private boolean paused=false;

	public OscilloscopePanel() {
		setPreferredSize(new Dimension(800,600));
		setBackground(new java.awt.Color(255, 255, 255));
		setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));	
	}

	@Override
	public void paintComponent(Graphics g){
		// Reset the display
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, getWidth(), getHeight());
		// Draw horizontal grid
		for(int i=1;i<Config.nbGridsY();i++){
			if(i==Config.nbGridsY()/2)
				g.setColor(Color.BLACK);
			else
				g.setColor(Color.GRAY);
			int row=(this.getHeight()*i)/Config.nbGridsY();
			g.drawLine(0, row, this.getWidth(), row);
		}

		// Draw vertical grid
		for(int i=1;i<Config.nbGridsX();i++) {
			if(i==Config.nbGridsX()/2)
				g.setColor(Color.BLACK);
			else
				g.setColor(Color.GRAY);
			int col=(getWidth()*i)/Config.nbGridsX();
			g.drawLine(col, 0, col, this.getHeight());
		}

		drawChart(g,buffer);
	}

	/**
	 * Draw the oscilloscope chart on the given Graphics
	 * @param g
	 * @param buffer
	 */
	private void drawChart(Graphics g, short[] buffer) {
		if(buffer==null) return;
		short max2=(short) ((Short.MAX_VALUE*threshold)/100);
		short min2=(short) ((Short.MIN_VALUE*threshold)/100);
		boolean up=false;
		int indexUp=-1;
		int indexDown=-1;

		int cross=0;
		for(int i=0;i<buffer.length;i++) {
			if(!up && buffer[i]>max2) { 
				up=true;
				// Synchronize on an hysteresis transition from low to high
				if(indexUp==-1 && cross > 1) indexUp=i;
				cross++;
			}
			if(up && buffer[i]<min2) {
				up=false;
				if(indexDown==-1 && cross > 1) indexDown=i;
			}
			if(indexDown!=-1 && indexUp!=-1)
				break;
		}

		// If up/down front is too far in the buffer, we failed to synchronized
		if(indexUp>(buffer.length/3)) indexUp=-1;
		if(indexDown>(buffer.length/3)) indexDown=-1;

		// Select UP/DOWN front
		if(synchro)
			index=indexUp;
		else 
			index=indexDown;

		float scaleBuffer=(float) ((scale*Config.defaultSampleRate())/(float)getWidth());

		if(paused) {
			g.setColor(Color.RED); // Paused chart is displayed as red
		} else {
			if(index<0) 
				g.setColor(Color.GRAY); // Unsynched chart is displayed as gray
			else
				g.setColor(Color.BLUE); // Synched chart is displayed as blue
		}
		if(index<0) index=0;

		if(buffer!=null) {
			float y1=this.getHeight()*(1.F-(float)(buffer[index%buffer.length]-min)/(float)(max-min));

			for(int i=1;i<this.getWidth();i++) {
				//int idx=(int) (i*scaleBuffer);
				//float val=getBufferValue(i*scaleBuffer,buffer);
				//float y2=this.getHeight()*(1.F-(float)(buffer[(index+idx)%buffer.length]-min)/(float)(max-min));
				float y2=this.getHeight()*(1.F-(float)(getBufferValue(index+i*scaleBuffer,buffer,scaleBuffer)-min)/(float)(max-min));
				g.drawLine(i-1, (int)y1, i, (int)y2);
				y1=y2;
			}
		}
	}

	private float getBufferValue(float idx, short[] bufferShort,float scaleBuffer) {
		if(scaleBuffer<=1)
			return bufferShort[((int)idx)%bufferShort.length];
		int idx2=(int)idx;
		short val1=bufferShort[idx2%bufferShort.length];
		short val2=bufferShort[(idx2+1)%bufferShort.length];
		float u=idx-idx2;
		return val1*(1-u)+val2*u;
	}

	/**
	 * @param buffer2
	 * @param min
	 * @param max
	 * @param index
	 * @param max2
	 * @param min2
	 * @param s
	 */
	public void setBuffer(short[] buffer2, short min,short max) {
		buffer=buffer2;
		this.min=min;
		this.max=max;
	}

	public void setScale(double d) {
		this.scale=d;
		repaint();
	}

	public void setThreshold(float d) {
		this.threshold=d;
	}

	public void setPause(boolean selected) {
		paused=selected;
	}

	public void setSynchro(boolean selected) {
		synchro = selected;
	}

	public boolean isPause() {
		return paused;
	}

}
