/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.color.ColorSpace;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.Arrays;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.BevelBorder;

/**
 * @author Gilles Bernard
 * Panel for displaying the spectrogram. X is time, Y is frequency and value is represented by a color scale.
 */
public class SpectrogramPanel extends JPanel implements MouseMotionListener {

	private static final long serialVersionUID = 6360887160294401903L;


	// Values for bandwidth and centralFreq. These values control the display of the spectrogram (height of the screen).
	private double bandwidth;
	private double centralFreq;
	// Label for frequency under the cursor. The label is stored in the main frame and passed to the panel.
	// TODO use another mean to pass the frequency value.
	private JLabel freqLabel;

	// Indicates whether the vertical scale is decimal (frequencies) or musical (notes).
	private boolean decimal=true;

	private int nbSamples =0;
	private double scale; // Duration for the panel width in second
	private BufferedImage[] tiles;
	private long startTime=0;
	private long pauseTime=0;
	private double scaleX;
	private double duration;
	private float resolution;
	private double fmin;
	private double fmax;
	private int idxMin;
	private double vertStep;
	private int idxMax;
	int lastIdx=0;
	private Vector<byte[]> lastFfts=new Vector<byte[]>();


	private float sampleRate=Config.defaultSampleRate();

	public SpectrogramPanel() {
		setPreferredSize(new Dimension(800,600));
		setBackground(new java.awt.Color(255, 255, 255));
		setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));	
		reset();
	}

	@Override
	public void paintComponent(Graphics g){
		// Reset the display
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, getWidth(), getHeight());
		try {
			if(tiles==null) return;

			// Find out in which tile we should draw
			long time=(Util.clock()-startTime);
			if(pauseTime>0) time=pauseTime-startTime;
			if(time<0) time=0;
			time%=duration;
			int idxTile=(int) ((time/duration)*tiles.length);
			int offset=(int) ((time-idxTile*Config.tileSize()*scaleX)/scaleX);

			// Draw as an horizontal waterfall
			for(int i=0;i<tiles.length;i++) {
				g.drawImage(tiles[(idxTile+i)%tiles.length],
						(getWidth()+i*Config.tileSize()-offset)%(tiles.length*Config.tileSize()),0,null);
			}

			// Cleanup left side for drawing scale values
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, 20, getHeight());

			// Draw the scales
			if(decimal) { // Decimal scale
				int prevScale=-10000;
				for(double scale : Config.timeScales()){
					double row = (1/scale-idxMin*resolution)/(vertStep*resolution);
					if(row>0 && row < getHeight() && (row > (prevScale+15))) {
						g.setColor(Color.GRAY);
						g.drawLine(20,getHeight()-(int)row, this.getWidth(), getHeight()-(int)row);
						g.setColor(Color.BLACK);
						g.drawString(Util.frequenceToText(1/scale), 2, getHeight()-(int)row);
						prevScale=(int)row;
					}
				}
			} else {  // Musical scale
				int prevScale=-10000;
				for (int i=0;i<Config.getNotesFrequencies().length;i++) {
					double col = (Config.getNotesFrequencies()[i]-idxMin*resolution)/(vertStep*resolution);
					if(col>0 && col < getHeight() && col > (prevScale+15)) {
						g.setColor(Color.GRAY);
						g.drawLine(20, getHeight()-(int)col, this.getWidth(), getHeight()-(int)col);
						g.setColor(Color.BLACK);
						g.drawString(Config.getNotesNames()[i], 2, getHeight()-(int)col);
						prevScale=(int)col;
					}
				}
			}

			// Draw vertical grid
			for(int i=1;i<Config.nbGridsX();i++) {
				if(i==Config.nbGridsX()/2)
					g.setColor(Color.GREEN);
				else
					g.setColor(Color.GRAY);
				int col=(getWidth()*i)/Config.nbGridsX();
				g.drawLine(col, 0, col, this.getHeight());
			}

		} catch (Exception e){
			e.printStackTrace();
		}
	}

	@Override
	public void mouseDragged(MouseEvent arg0) {

	}

	@Override
	public void mouseMoved(MouseEvent arg0) {

		double f = idxMin * resolution +(getHeight()-arg0.getY())*vertStep*resolution;

		freqLabel.setText(Util.frequenceToText(f));
	}

	/**
	 * Set the bandwidth. 
	 * The display scales from central frequency +/- bandwidth/2. 
	 * @param d
	 */
	public synchronized void setBandwidth(double d) {
		this.bandwidth=d;
		reset();
	}

	/**
	 * Set the central frequency. 
	 * The display scales from central frequency +/- bandwidth/2. 
	 * @param d
	 */
	public synchronized void setCentralFrequency(double d) {
		this.centralFreq=d;
		reset();
	}

	/**
	 * @param freqLabel
	 * Set the label to be used to display the frequency under the cursor.
	 */
	public void setFrequenceLabel(JLabel freqLabel) {
		this.freqLabel=freqLabel;
	}

	/**
	 * Set decimal scale (true) or musical scale (false).
	 * @param selected
	 */
	public void setDecimal(boolean selected) {
		this.decimal=selected;
	}

	/**
	 * Set the current fft as a byte array. Values in this array scales from 0 to Byte.MAX_VALUE.
	 * @param fftByte
	 */
	public void setBuffer(byte[] fftByte) {
		synchronized (this) {
			// There was a change in the number of samples, take this new value and reset everything.
			if(fftByte.length!=nbSamples ) {
				nbSamples=fftByte.length;
				reset();
			}

			synchronized (lastFfts) {
				// Find out in which tile we should draw
				//long time=(System.currentTimeMillis()-startTime);
				long timeCur=Util.clock();
				if(pauseTime>0) timeCur=pauseTime;
				long time=(timeCur-startTime)%(long)duration;
				int idxTile=(int) ((time/duration)*tiles.length);
				int offset=(int) (((double)time-(double)idxTile*(double)Config.tileSize()*scaleX)/scaleX);

				int idx=idxTile*Config.tileSize()+offset;
				if(lastIdx==idx && lastIdx>0) {
					// Same Idx, memorize arrays because of timer inaccuracy
					lastFfts.add(Arrays.copyOf(fftByte, fftByte.length));
					//System.out.println("Add "+lastIdx);
				} else {

					// Spread memorized ffts from lastIdx to idx (several ffts arrived with the same idx due to timer inaccuracy)
					if(lastFfts.size()>0) {
						//System.out.println("Proceed "+lastIdx+" -> "+idx+" "+lastFfts.size());
						lastFfts.add(fftByte);
						for(int i=lastIdx+1;i<=idx;i++){
							int idx2=(i/Config.tileSize());
							int off2=i%Config.tileSize();
							//System.out.println("R "+idx2+" "+off2);
							Graphics gc=tiles[idx2%tiles.length].getGraphics();
							if(idx>(lastIdx+1))
								drawSpectrum(gc,off2,lastFfts.elementAt(((i-lastIdx-1)*(lastFfts.size()-1))/(idx-lastIdx-1)));
							else
								drawSpectrum(gc,off2,lastFfts.elementAt(0));
						}
						lastIdx=idx;
						lastFfts.clear();
					} else { 
						// Spread current fft from lastIdx to idx (due to timer inaccuracy).
						for(int i=lastIdx;i<=idx;i++) {
							int idx2=(i/Config.tileSize());
							int off2=i%Config.tileSize();
							Graphics gc=tiles[idx2%tiles.length].getGraphics();
							drawSpectrum(gc,off2,fftByte);
						}
					}

					lastIdx=idx;
				}
			}
		}
	}

	/**
	 * Draw the given fft at the given offset (X) in the graphics g.
	 * @param g
	 * @param off
	 * @param fftByte
	 */
	private void drawSpectrum(Graphics g, int off, byte[] fftByte) {
		synchronized(this) {
			for(int y=0;y<getHeight();y++) {
				byte val=fftByte[idxMin+(int)(vertStep*y)];
				Color c=Util.valueToColor(val);
				g.setColor(c);
				g.drawLine(off,getHeight()-y,off,getHeight()-y);
			}
		} 
	}

	/**
	 * Reset the display. 
	 */
	public void reset() {
		if(getHeight()>0 && getWidth()>0) {
			synchronized (lastFfts) {
				// Called when frame changes size
				int nbTiles=getWidth()/Config.tileSize();
				nbTiles+=2;
				this.tiles=new BufferedImage[nbTiles];
				for(int i=0;i<nbTiles;i++) {
					tiles[i]=new BufferedImage(Config.tileSize(),getHeight(),ColorSpace.TYPE_RGB);
					Graphics gc=tiles[i].getGraphics();
					gc.setColor(Color.BLACK);
					gc.fillRect(0, 0, Config.tileSize(), getHeight());
				}

				startTime=Util.clock();

				lastIdx=0;
				
				lastFfts.clear();
				pauseTime=0;
				// s/px duration of a screen pixel
				scaleX=scale*Util.clock_to_second()/getWidth();
				// s duration of the whole buffer of tiles
				duration=tiles.length*Config.tileSize()*scaleX;
				// duration of the acquisition of nbSamples
				resolution=sampleRate/nbSamples;
				// frequency min at the bottom of the screen
				fmin=centralFreq-bandwidth/2;
				if(fmin<0) fmin=0;
				// frequency max at the top of the screen
				fmax=centralFreq+bandwidth/2;
				idxMin=(int) (fmin/resolution); // idx corresponding to fmin
				idxMax=(int) (fmax/resolution); // idx corresponding to fmax
				// vertical resolution
				vertStep=((double)(idxMax-idxMin))/(double)getHeight(); 
			}
		}
	}


	/**
	 * Set the vertical scale. The scale is the number of second for the whole panel width.
	 * @param d
	 */
	public void setScale(double d) {
		this.scale=d;
		reset();
	}

	/**
	 * Called whenever the pause is selected/unselected.
	 * Restores lastIdx to the current value to keep the display consistent.
	 * @param selected
	 */
	public void setPause(boolean selected) {
		if(!selected) {
			long timeCur=Util.clock();
			startTime+=timeCur-pauseTime;
			long time=(timeCur-startTime)%(long)duration;
			int idxTile=(int) ((time/duration)*tiles.length);
			int offset=(int) (((double)time-(double)idxTile*(double)Config.tileSize()*scaleX)/scaleX);

			lastIdx=idxTile*Config.tileSize()+offset;
			pauseTime=0;
			if(lastIdx<0) reset();
		} else {
			pauseTime=Util.clock();
		}
	}

	public void setSampleRate(float f) {
		sampleRate=f;
	}


}
