/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Vector;

import javax.imageio.ImageIO;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * @author Gilles Bernard
 * Main class of the Spectrum application. Spectrum is displayed frequency in X and value in Y.
 */
public class Spectrum extends JFrame implements ChangeListener, ActionListener {

	private static final long serialVersionUID = 5312114580592842412L;
	private static int nbSamples = 1<<12;
	private SpectrumPanel panel;
	private AudioThread audioThread=null;
	private JPanel command;
	private JSlider bandwidthSlider;
	private JLabel bandwitdthLabel;
	private JCheckBox memCheckBox;
	private VolumePanel volumePanel;
	private JSlider centralFreqSlider;
	private JLabel centralFreqLabel;
	private JLabel freqLabel;
	private JRadioButton decimalRadio;
	private JRadioButton musicRadio;
	private JSlider nbSampleSlider;
	private JLabel nbSampleLabel;
	private JCheckBox noiseCheckBox;
	private long startNoise=0;
	private float[] fftNoise;
	private int nbFftsNoise;
	private boolean noiseSet;
	private JSlider nbSampleSliderAcq;
	private JLabel nbSampleLabelAcq;
	private double[] volumeBuf=new double[1];
	private int volumeIdx=0;

	/**
	 * 
	 */
	public Spectrum() {
		setTitle("Lordikc AudioAnalyst Spectrum");
		InputStream is = getClass().getResourceAsStream("/oscillo_128.png");
		try {
			if(is!=null)
				this.setIconImage(ImageIO.read(is));
			else System.out.println("Can't load icon");
		} catch (IOException e) {
			e.printStackTrace();
		}
		setLayout(new BorderLayout());
		// The panel display the chart
		panel = new SpectrumPanel();
		panel.addMouseMotionListener(panel);
		getContentPane().add(panel,BorderLayout.CENTER);
		// The command panel
		command = new JPanel();
		getContentPane().add(command,BorderLayout.EAST);
		command.add(Box.createRigidArea(new Dimension(0,5)));
		command.setLayout(new BoxLayout(command, BoxLayout.PAGE_AXIS));

		// Bandwidth label, slider and value
		JLabel label=new JLabel("Bandwidth");
		command.add(label);
		bandwidthSlider = new JSlider(JSlider.HORIZONTAL,
				0, Config.timeScalesBw().length-1, Config.defaultBandwidth());
		bandwidthSlider.addChangeListener(this);
		bandwidthSlider.setPreferredSize(new Dimension(100,20));
		command.add(bandwidthSlider);
		bandwitdthLabel=new JLabel();
		command.add(bandwitdthLabel);
		setBandwidth(Spectrum.sliderToFrequenceBw(bandwidthSlider));

		// Central frequency label, slider and value
		label=new JLabel("Cent. Freq.");
		command.add(label);
		centralFreqSlider = new JSlider(JSlider.HORIZONTAL,
				0, Config.timeScales().length-1, Config.defaultCentralFrequency());
		centralFreqSlider.addChangeListener(this);
		centralFreqSlider.setPreferredSize(new Dimension(100,20));
		command.add(centralFreqSlider);
		centralFreqLabel=new JLabel();
		command.add(centralFreqLabel);
		setCentralFrequency(Spectrum.sliderToFrequence(centralFreqSlider));

		// Nb samples per second label, slider and value
		nbSampleSliderAcq = new JSlider(JSlider.HORIZONTAL,
				0, Config.sampleRates().length-1, Config.sampleRates().length-1);
		nbSampleSliderAcq.addChangeListener(this);
		nbSampleSliderAcq.setPreferredSize(new Dimension(100,20));
		label=new JLabel("Samples/sec");
		command.add(label);
		command.add(nbSampleSliderAcq);
		nbSampleLabelAcq=new JLabel();
		command.add(nbSampleLabelAcq);

		// Nb samples per second label, slider and value
		label=new JLabel("Samples/scan");
		command.add(label);
		nbSampleSlider = new JSlider(JSlider.HORIZONTAL,
				Config.minSamples(), Config.maxSamples(), Config.defaultSamples());
		nbSampleSlider.addChangeListener(this);
		nbSampleSlider.setPreferredSize(new Dimension(100,20));
		command.add(nbSampleSlider);
		nbSampleLabel=new JLabel();
		command.add(nbSampleLabel);

		// Checkbox for holding the display
		memCheckBox=new JCheckBox("Hold");
		memCheckBox.setActionCommand(Config.HOLD);
		memCheckBox.addActionListener(this);
		command.add(memCheckBox);

		// Checkbox for noise reduction
		noiseCheckBox=new JCheckBox("Noise");
		noiseCheckBox.setActionCommand(Config.NOISE);
		noiseCheckBox.addActionListener(this);
		command.add(noiseCheckBox);

		// Radio buttons for selecting musical or decimal grid
		decimalRadio = new JRadioButton("Dec.");
		decimalRadio.setActionCommand(Config.DECIMAL_MUSIC_SCALE);
		decimalRadio.addActionListener(this);
		decimalRadio.setSelected(true);
		musicRadio = new JRadioButton("Mus.");
		musicRadio.setActionCommand(Config.DECIMAL_MUSIC_SCALE);
		musicRadio.addActionListener(this);
		ButtonGroup scaleMode = new ButtonGroup();
		scaleMode.add(decimalRadio);
		scaleMode.add(musicRadio);
		command.add(decimalRadio);
		command.add(musicRadio);

		// Label for display the frequency under the pointer
		command.add(new JLabel("Freq"));
		freqLabel=new JLabel("0Hz");
		command.add(freqLabel);
		panel.setFrequenceLabel(freqLabel);

		// About button
		JButton button=new JButton("About");
		button.setActionCommand("about");
		button.addActionListener(this);
		command.add(button);

		// Panel displaying the volume as a bar chart
		volumePanel=new VolumePanel();
		Box bv = Box.createVerticalBox();
		bv.add(Box.createVerticalGlue());
		bv.add(volumePanel);
		bv.add(Box.createVerticalGlue());
		Box bh = Box.createVerticalBox();
		bh.add(Box.createHorizontalGlue());
		bh.add(bv);
		bh.add(Box.createHorizontalGlue());
		command.add(bh);

		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		pack();
	}

	/**
	 * 
	 */
	public void init() {
		setNbSamples(1<<nbSampleSlider.getValue());
		setSampleRate(Config.sampleRates().length-1);
	}
	
	private void setSampleRate(int i) {
		nbSampleLabelAcq.setText(Integer.toString((int) Config.sampleRates()[i]));
		panel.setSampleRate(Config.sampleRates()[i]);
		audioThread.setSampleRate(Config.sampleRates()[i]);
		panel.repaint();
		reset_volume();
	}

	/**
	 * Sets the number of sample to be read out of the soundcard. 
	 * It changes the audio thread settings along with the label.
	 * @param i
	 */
	private void setNbSamples(int i) {
		if(audioThread!=null) {
			audioThread.setNbSamples(i);
			nbSampleLabel.setText(Integer.toString(i));
		}
		reset_noise();
		reset_volume();
	}

	/**
	 * 
	 */
	private void reset_noise() {
		noiseSet=false;
		synchronized (this) {fftNoise=null;}
		nbFftsNoise=0;
		noiseCheckBox.setSelected(false);
		startNoise=0;
	}

	/**
	 * 
	 */
	private void reset_volume() {
		if(audioThread==null)
			return;
		int n=(int) (audioThread.getSampleRate()/audioThread.getNbSamples());
		if(n==0) n=1;
		volumeBuf = new double[n];
		volumeIdx=0;
	}

	private static double sliderToFrequence(JSlider slider) {
		return 1/Config.timeScales()[slider.getValue()];
	}

	private static double sliderToFrequenceBw(JSlider slider) {
		return 1/Config.timeScalesBw()[slider.getValue()];
	}

	private void setBandwidth(double d) {
		panel.setBandwidth(d);
		bandwitdthLabel.setText(Double.toString(d)+"Hz");
	}

	private void setCentralFrequency(double d) {
		panel.setCentralFrequency(d);
		centralFreqLabel.setText(Double.toString(d)+"Hz");
	}

	/**
	 * @param args
	 * @throws Exception
	 */
	public static void main(String args[]) throws Exception {

		// Create the main frame of the application
		final Spectrum frame = new Spectrum();
		// Create the recording and processing thread.
		SpectrumProcessThread process = new SpectrumProcessThread(nbSamples,frame);
		Vector<ISampleListener> listeners = new Vector<ISampleListener>();
		listeners.add(process);
		AudioThread audio = new AudioThread(listeners,nbSamples);
		frame.setAudioThread(audio);
		frame.init();

		// Do display the GUI
		EventQueue.invokeLater(new Runnable() {
			public void run() {
				frame.setVisible(true);
			}
		});

		// Start threads
		process.start();
		audio.start();

		// Wait for the end
		audio.join();
		process.join();
	}

	/**
	 * @param audio
	 */
	private void setAudioThread(AudioThread audio) {
		this.audioThread=audio;
		reset_volume();
	}

	@Override
	public void stateChanged(ChangeEvent event) {
		// React to changes in bandwidth slider
		if(event.getSource().equals(bandwidthSlider)) {
			setBandwidth(Spectrum.sliderToFrequenceBw(bandwidthSlider));
			panel.repaint();
		}
		// React to changes in central freq slider
		if(event.getSource().equals(centralFreqSlider)) {
			setCentralFrequency(Spectrum.sliderToFrequence(centralFreqSlider));
			panel.repaint();
		}
		// React to change in number of sample slider
		if(event.getSource().equals(nbSampleSlider)) {
			setNbSamples(1<<nbSampleSlider.getValue());
			panel.repaint();
		}
		// React to sample rate slider changes
		if(event.getSource().equals(nbSampleSliderAcq)) {
			setSampleRate(nbSampleSliderAcq.getValue());
			panel.repaint();
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		if(event.getActionCommand().equals(Config.HOLD)){
			// React to changes in HOLD CheckBox
			audioThread.setPause(memCheckBox.isSelected());
			panel.repaint();
		} else if(event.getActionCommand().equals(Config.DECIMAL_MUSIC_SCALE)){
			// React to changes in decimal/musical button radio 
			panel.setDecimal(decimalRadio.isSelected());
			panel.repaint();
		} else if(event.getActionCommand().equals(Config.NOISE)){
			// React to changes in noise checkbox 
			if(noiseCheckBox.isSelected()) {
				startNoise=Util.clock();
				synchronized (this) {fftNoise=null;}
				nbFftsNoise=0;
			} else {
				startNoise=0;
			}
			noiseSet=false;
		} else if(event.getActionCommand().equals(Config.ABOUT)){
			// React to click on about button.
			Util.about(this);
		}
	}

	/**
	 * This method is called by the processing thread to set the fft.
	 * @param fftRe
	 * @param max
	 */
	public synchronized void setBuffer(float[] fftRe, double max) {
		// Allocate the buffer for storing the average noise 
		if(fftNoise==null || fftNoise.length!=fftRe.length) {
			fftNoise=new float[fftRe.length];
			Arrays.fill(fftNoise, 0);
			nbFftsNoise=0;
			noiseSet=false;
		}
		// Compute the average noise from timeMinNoise to timeMaxNoise
		if(startNoise>0 && (Util.clock()-startNoise)>(Config.timeMinNoise()*Util.clock_to_second())) {
			for(int i=0;i<fftRe.length;i++)
				fftNoise[i]+=fftRe[i];
			nbFftsNoise++;
			if((Util.clock()-startNoise)>(Config.timeMaxNoise()*Util.clock_to_second())) {
				for(int i=0;i<fftNoise.length;i++) {
					fftNoise[i]/=nbFftsNoise;
				}
				// timeMaxNoise is reached indicate that noise is set
				noiseSet=true;
				startNoise=0;
			}
		}

		// If noise is set, substract the average noise from the current buffer.
		if(noiseSet && fftNoise.length==fftRe.length) {
			max=-Double.MAX_VALUE;
			for(int i=0;i<fftNoise.length;i++) {
				fftRe[i]-=fftNoise[i];
				if(fftRe[i]<0) fftRe[i]=0;
				if(fftRe[i]>max) max = fftRe[i];
			}
		}

		volumeBuf[volumeIdx++%volumeBuf.length]=max;
		
		// Forward the fft to the panel.
		panel.setBuffer(fftRe, volumeIdx>volumeBuf.length?Util.mean(volumeBuf):max);
		panel.repaint();
	}

	/**
	 * @param vol
	 */
	public void setVolume(short vol) {
		volumePanel.setVolume(vol);
		volumePanel.repaint();
	}
}

