/**
 *   This file is part of AudioAnalyst.
 *
 *   AudioAnalyst is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   AudioAnalyst is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.lordikc.audioanalyst;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.BevelBorder;

/**
 * @author Gilles Bernard
 * Main panel of the spectrum application.
 */
public class SpectrumPanel extends JPanel implements MouseMotionListener {


	private static final long serialVersionUID = 6360887160294401903L;

	private float[] buffer=null;
	private double max;

	private double bandwidth;
	private double centralFreq;
	private JLabel freqLabel;
	private boolean decimal=true;

	private float sampleRate = Config.defaultSampleRate();

	/**
	 * 
	 */
	public SpectrumPanel() {
		setPreferredSize(new Dimension(800,600));
		setBackground(new java.awt.Color(255, 255, 255));
		setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));	
	}

	@Override
	public void paintComponent(Graphics g){
		// Reset the display
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, getWidth(), getHeight());
		try {
			if(buffer==null || buffer.length==0) return;
			g.setColor(Color.BLACK);
			g.drawLine(0, this.getHeight()/2, this.getWidth(), this.getHeight()/2);
			g.drawLine(0, this.getHeight()/4, this.getWidth(), this.getHeight()/4);
			g.drawLine(0, (3*this.getHeight())/4, this.getWidth(), (3*this.getHeight())/4);

			double resolution=sampleRate/buffer.length;

			// f = idx_buf * resolution
			// fmin = fcent - bw/2 -> i_win = 0
			// fmax = fcent + bw/2 -> i_win = getWidth()
			// idx_min=fmin/resolution
			// idx_max=fmax/resolution

			double fmin=centralFreq-bandwidth/2;
			if(fmin<0) fmin=0;
			double fmax=centralFreq+bandwidth/2;
			int idxMin=(int) (fmin/resolution);
			int idxMax=(int) (fmax/resolution);
			double step=((double)(idxMax-idxMin))/(double)getWidth();
			// f = idx_buf * resolution
			// idx_buf = idxMin+step*idx_ecr
			// f = idxMin * resolution +idx_ecr*step*resolution
			// idx_ecr = (f-idxMin*resolution)/(step*resolution)

			// Draw decimal or musical scale vertically
			if(decimal) {
				int scaleY=0;
				int prevScale=-1000;
				for(double scale : Config.timeScales()){
					double col = (1/scale-idxMin*resolution)/(step*resolution);
					if(col>0 && col < getWidth() && (col > (prevScale+15))) {
						g.setColor(Color.GRAY);
						g.drawLine((int)col, 20, (int)col, this.getHeight());
						g.setColor(Color.BLACK);
						g.drawString(Util.frequenceToText(1/scale), (int)col+2, 10+10*(scaleY++%2));
						prevScale=(int)col;
					}
				}
			} else {
				int scaleY=0;
				int prevScale=-1000;
				for (int i=0;i<Config.getNotesFrequencies().length;i++) {
					double col = (Config.getNotesFrequencies()[i]-idxMin*resolution)/(step*resolution);
					if(col>0 && col < getWidth() && col > (prevScale+15)) {
						g.setColor(Color.GRAY);
						g.drawLine((int)col, 0, (int)col, this.getHeight());
						g.setColor(Color.BLACK);
						g.drawString(Config.getNotesNames()[i], (int)col+2, 10+10*(scaleY++%2));
						prevScale=(int)col;
					}
				}
			}

			// Draw the spectrum
			g.setColor(Color.BLUE);
			for(int i=0;i<this.getWidth();i++) {
				double idx=step*i;
				double val=buffer[idxMin+(int)idx]/max;
				double y2=20+(this.getHeight()-20)*(1.-val);
				g.drawLine(i, this.getHeight(), i, (int)y2);
			}
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	@Override
	public void mouseDragged(MouseEvent arg0) {
		// Ignored events
	}

	@Override
	public void mouseMoved(MouseEvent arg0) {
		if(buffer==null || buffer.length==0) return;
		double resolution=sampleRate/buffer.length;

		// f = idx_buf * resolution
		// fmin = fcent - bw/2 -> i_win = 0
		// fmax = fcent + bw/2 -> i_win = getWidth()
		// idx_min=fmin/resolution
		// idx_max=fmax/resolution

		// Compute the frequency under the pointer
		double fmin=centralFreq-bandwidth/2;
		if(fmin<0) fmin=0;
		double fmax=centralFreq+bandwidth/2;
		int idxMin=(int) (fmin/resolution);
		int idxMax=(int) (fmax/resolution);
		double step=((double)(idxMax-idxMin))/(double)getWidth();
		// f = idx_buf * resolution
		// idx_buf = idxMin+step*idx_ecr
		double f = idxMin * resolution +arg0.getX()*step*resolution;
		// idx_ecr = (f-idxMin*resolution)/(step*resolution)

		// Display it
		freqLabel.setText(Util.frequenceToText(f)+"Hz");
	}

	/**
	 * Set bandwidth of the display. Panel horizontally ranges from central freq -/+ bandwidth/2
	 * @param d
	 */
	public void setBandwidth(double d) {
		this.bandwidth=d;
	}

	/**
	 * Set central frequency of the display. Panel horizontally ranges from central freq -/+ bandwidth/2
	 * @param d
	 */
	public void setCentralFrequency(double d) {
		this.centralFreq=d;
	}

	public synchronized void setBuffer(float[] fftRe, double max2) {
		this.buffer=fftRe;
		this.max=max2;
	}

	/**
	 * Sets the frequency under the pointer in the label held by the frame.
	 * @param freqLabel
	 */
	public void setFrequenceLabel(JLabel freqLabel) {
		this.freqLabel=freqLabel;
	}

	/**
	 * Set the scale decimal/musical
	 * @param selected
	 */
	public void setDecimal(boolean selected) {
		this.decimal=selected;
	}

	public void setSampleRate(float f) {
		sampleRate =f;
	}

}
